#!/usr/bin/env python

"""
Simple indexing of sorted files.

Copyright (C) 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

Motivations
-----------

When searching files, the cost to scan an entire file typically exceeds the
cost to navigate to the appropriate region and to scan a smaller portion of the
file, especially for large files. At the same time, exotic data structures
encouraging multiple seeks and reads are likely to waste time compared to just
performing a single read operation, even if that operation involves a larger
quantity of data, at least for storage with hard disk access characteristics.
"""

from simplex.readers import *
from simplex.indexers import *
import bisect

class IndexedFile:

    "An indexed file referring to a sorted file."

    def __init__(self, reader, index, get_key):
        self.reader = reader
        self.index = index
        self.get_key = get_key

    def find(self, term):
        return find_with_index(self.reader, self.get_key, self.index, term)

def find_with_index(reader, get_key, l, term):

    """
    In the resource whose 'reader' provides records, using a 'get_key' operation
    to yield the key for such records, and using the given index list 'l', find
    the given 'term', returning a record employing the term or None if no such
    record was found.
    """

    i = bisect.bisect_left(l, (term, None))

    try:
        found, pos = l[i]
    except IndexError:
        return None

    # Since the index is more coarse than the underlying file, the bisect left
    # operation will most likely point to an index entry for later records than
    # the desired one.

    if found > term:
        i = max(0, i - 1)
        found, pos = l[i]

    reader.seek(pos)

    if found == term:
        return reader.next()
    else:
        return find_in_file(reader, get_key, term)

def find_in_file(reader, get_key, term):

    """
    In the resource whose 'reader' provides records, using a 'get_key' operation
    to yield the key for such records, find the given 'term', returning a record
    employing the term or None if no such record was found.
    """

    for record in reader:
        key = get_key(record)
        if term == key:
            return record

        # Short-circuit failed searches.

        elif term < key:
            return None

    return None

def groups(l, length):

    "Split 'l' into groups of the given 'length'."

    if length <= 0:
        raise ValueError, "Groups must be greater than zero."

    i = 0
    g = []

    while i < len(l):
        g.append(l[i:i+length])
        i += length

    return g

# vim: tabstop=4 expandtab shiftwidth=4
